/*
 * config.c:
 * Read tlsproxyd config.
 *
 * Copyright (c) 2002 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: config.c,v 1.2 2002/06/11 13:52:19 chris Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#include "tlsproxyd.h"

/* facil:
 * Log facility names and constants, used to allow configurable logging at
 * run time. Not all those defined in openlog(3) make sense, so only a few
 * are listed here. */
static struct logfac {
    char *name;
    int fac;
} facil[] = {
        {"mail",        LOG_MAIL},
#ifdef LOG_AUTHPRIV
        {"authpriv",    LOG_AUTHPRIV},
#endif
#ifdef LOG_AUTH
        {"auth",        LOG_AUTH},
#endif
        {"daemon",      LOG_DAEMON},
        {"user",        LOG_USER},
        
        {"local0",      LOG_LOCAL0},
        {"local1",      LOG_LOCAL1},
        {"local2",      LOG_LOCAL2},
        {"local3",      LOG_LOCAL3},
        {"local4",      LOG_LOCAL4},
        {"local5",      LOG_LOCAL5},
        {"local6",      LOG_LOCAL6},
        {"local7",      LOG_LOCAL7},
    };

#define NFACIL      (sizeof(facil) / sizeof(struct logfac))



/* read_config:
 * Read configuration from specified file. */
struct config *read_config(const char *filename) {
    FILE *fp;
    char linebuf[1024];
    int line = 1;
    struct config *cfg;
    int have_logfacil;

    cfg = xcalloc(sizeof *cfg, 1);

    fp = fopen(filename, "rt");
    if (!fp) {
        log_print(LOG_ERR, "%s: %m", filename);
        return NULL;
    }
    
    while (fgets(linebuf, sizeof linebuf, fp) != NULL) {
        char *p;
        char b1[1024], b2[1024], b3[1024], b4[1024];

        if (linebuf[strlen(linebuf) - 1] != '\n') {
            log_print(LOG_WARNING, "%s:%d: ignoring overlong line", linebuf, line);
            continue;
        }
        linebuf[strlen(linebuf) - 1] = 0;
        
        /* Comments. */
        p = strchr(linebuf, '#');
        if (p)
            *p = 0;

        p = linebuf + strspn(linebuf, " \t");

        if (!*p)
            /* Blank line. */
            continue;
        else if (sscanf(p, "%[^ =] = %[^ =]", b1, b2) == 2) {
            /* foo = bar directive */
            if (!strcmp(b1, "log-facility")) {
                if (have_logfacil) {
                    log_print(LOG_WARNING, "%s:%d: repeated log-facility directive ignored", filename, line);
                } else {
                    int i;
                    for (i = 0; i < NFACIL; ++i) {
                        if (!strcmp(b2, facil[i].name)) {
                           cfg->logfacil = facil[i].fac;
                           have_logfacil = 1;
                        }
                    }
                    if (!have_logfacil) {
                        log_print(LOG_WARNING, "%s:%s: `%s' is not a known log facility; using `daemon'", filename, line, b2);
                        cfg->logfacil = LOG_DAEMON;
                        have_logfacil = 1;
                    }
                }
            } else if (!strcmp(b1, "certificate")) {
                if (cfg->certfile)
                    log_print(LOG_WARNING, "%s:%d: repeated certificate directive ignored", filename, line);
                else
                    cfg->certfile = strdup(b2);
            } else if (!strcmp(b1, "private-key")) {
                if (cfg->pkeyfile)
                    log_print(LOG_WARNING, "%s:%d: repeated private-key directive ignored", filename, line);
                else
                    cfg->pkeyfile = strdup(b2);
            } else if (!strcmp(b1, "pid-file")) {
                if (cfg->pidfile)
                    log_print(LOG_WARNING, "%s:%d: repeated pid-file directive ignored", filename, line);
                else
                    cfg->pidfile = strdup(b2);
            } else if (!strcmp(b1, "max-processes")) {
                if (cfg->maxprocs)
                    log_print(LOG_WARNING, "%s:%d: repeated max-processes directive ignored");
                else {
                    cfg->maxprocs = atoi(b2);
                    if (!cfg->maxprocs) {
                        log_print(LOG_ERR, "%s:%d: value `%s' for max-processes does not make sense", filename, line, b2);
                        goto fail;
                    }
                }
            } else if (!strcmp(b1, "timeout")) {
                if (cfg->timeout)
                    log_print(LOG_WARNING, "%s:%d: repeated timeout directive ignored", filename, line);
                else {
                    cfg->timeout = atoi(b2);
                    if (!cfg->timeout) {
                        log_print(LOG_ERR, "%s:%d: value `%s' for timeout does not make sense", filename, line, b2);
                        goto fail;
                    }
                }
            } else if (!strcmp(b1, "user")) {
                if (cfg->uid)
                    log_print(LOG_WARNING, "%s:%d: repeated user directive ignored");
                else if (!strtouid(b2, &cfg->uid)) {
                    log_print(LOG_ERR, "%s:%d: value `%s' for user does not make sense", filename, line, b2);
                    goto fail;
                } else if (!cfg->uid) {
                    log_print(LOG_ERR, "%s:%d: tlsproxyd must not be run as root", filename, line);
                    goto fail;
                }
            } else if (!strcmp(b1, "group")) {
                if (cfg->gid)
                    log_print(LOG_WARNING, "%s:%d: repeated group directive ignored");
                else if (!strtogid(b2, &cfg->gid)) {
                    log_print(LOG_ERR, "%s:%d: value `%s' for group does not make sense", filename, line, b2);
                    goto fail;
                } else if (!cfg->gid) {
                    log_print(LOG_ERR, "%s:%d: tlsproxyd must not be run as group 0", filename, line);
                    goto fail;
                }
            }
        } else if (sscanf(p, "%[^ :]:%[^ ] -> %[^ :]:%[^ ]", b1, b2, b3, b4) == 4) {
            /* host:port -> host:port */
            struct proxyent pe = {0};
            if (!strtoaddr(b1, &pe.from.sin_addr)) {
                log_print(LOG_ERR, "%s:%d: %s: host not found", filename, line, b1);
                continue;
            } else if (!strtoport(b2, &pe.from.sin_port)) {
                log_print(LOG_ERR, "%s:%d: %s: invalid port", filename, line, b2);
                continue;
            } else if (!strtoaddr(b3, &pe.to.sin_addr)) {
                log_print(LOG_ERR, "%s:%d: %s: host not found", filename, line, b3);
                continue;
            } else if (!strtoport(b4, &pe.to.sin_port)) {
                log_print(LOG_ERR, "%s:%d: %s: invalid port", filename, line, b4);
                continue;
            }

            pe.from.sin_family = pe.to.sin_family = AF_INET;
            
            /* OK, have a valid proxy spec; save it. */
            if ((cfg->nproxy & (cfg->nproxy - 1)) == 0)
                cfg->proxy = xrealloc(cfg->proxy, 2 * (cfg->nproxy + 1) * sizeof *cfg->proxy);
            cfg->proxy[cfg->nproxy++] = pe;
            
        } else {
            /* unknown */
            log_print(LOG_WARNING, "%s:%d: syntax error", filename, line);
        }

        ++line;
    }

    fclose(fp);

    if (!have_logfacil)
        cfg->logfacil = LOG_DAEMON;
    
#define IFMISSING(f, n) do if (!cfg->f) { \
                            log_print(LOG_ERR, "%s: required " n " directive missing", filename); \
                            goto fail; \
                        } while (0)
    IFMISSING(certfile, "certificate");
    IFMISSING(maxprocs, "max-processes");
    IFMISSING(uid, "user");
    IFMISSING(gid, "group");
#undef IFMISSING

    /* If no private key file has been specified, assume that it lives in the
     * same file as the certificate. */
    if (!cfg->pkeyfile) cfg->pkeyfile = cfg->certfile;

    /* If no timeout has been specified, set it to 60s. */
    if (!cfg->timeout) cfg->timeout = 60;
    
    if (!cfg->nproxy) {
        log_print(LOG_ERR, "%s: no proxy specifications given", filename);
        goto fail;
    } else
        return cfg;

fail:
    xfree(cfg->proxy);
    xfree(cfg);
    return NULL;
}
